/* CPU load monitor, using a pulse-width modulated disposable on-battery tester.
   (commonly found on Duracell or Energizer AA batteries).

   This program works under Linux and drives the display using a serial port's
   RTS line. Check the supplied schematic to build your own Duracell display.
   It is only a toy, it simply demonstrates how to drive such a display. It is
   unpolished, not robust at all, and should not be used in any sort of serious
   application (but you knew that didn't you?).

   You must have /dev/rtc support in the Linux kernel, and you will need to
   allow userland programs to request 1024 interrupts per second from the rtc.
   You can type the following command as root to achieve this:

   echo 1024 > /proc/sys/dev/rtc/max-user-freq

   Of course, you also need proper permissions to access the /dev/ttyS? device
   file corresponding to the serial port you connected the display too.

   Don't forget to calibrate the display with the -c switch before using the
   program. Also, be patient when you use the display, especially when it's
   cold, the "refresh rate" is quite low, in the order of 15s or more for the
   bar to stabilize.

   This program is distributed under the terms of the GNU General Public License
   See the COPYING file for details

   (c) P.P. Coupard / 2004
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <linux/rtc.h>


/* Definitions */
#define RTC_RATE			1024	/* per second */
#define CALIBRATION_FILE		"./duracell.cal"


/* Global variables */
int shmid;
int fdser;
int fdrtc;
int *pulse_width;
pid_t iopid;


/* Prototypes */
int io_process(void);


/* Main program */
int main(int argc,char **argv)
{
  int calibrate=0;
  int calibration_points[5];
  int calibration_table[100];
  FILE *calfd;
  FILE *procstatfd;
  char c;
  long l1,l2,l3,l4,prev_jiffies;
  char buf[256];
  int i,j;

  /* Test arguments */
  if(argc==3 && (!strcmp(argv[2],"-c") || !strcmp(argv[2],"--calibrate")))
  {
    calibrate=1;
    printf("Calibration mode\n");
    printf("----------------\n");
  }

  if(argc<2 || argc>2+calibrate)
  {
    printf("Usage: %s </dev/ttySx> [-c|--calibrate]\n",argv[0]);
    return(-1);
  }

  /* Allocate a shared memory block for the pulse width */
  if((shmid=shmget(rand(),sizeof(int),IPC_CREAT | 0600))==-1)
  {
    printf("Cannot allocate %d shared memory bytes\n",sizeof(int));
    return(-1);
  }

  /* Attach to the shared memory block */
  if((pulse_width=(int *)shmat(shmid,NULL,0))==(int *)-1)
  {
    shmctl(shmid,IPC_RMID,0);
    printf("Cannot attach to the shared memory block\n");
    return(-1);
  }

  /* Open the serial port */
  if((fdser=open(argv[1],O_WRONLY))==-1)
  {
    shmdt(pulse_width);
    shmctl(shmid,IPC_RMID,0);
    printf("Cannot open serial device %s\n",argv[1]);
    return(-1);
  }

  /* Open /dev/rtc */
  if((fdrtc=open("/dev/rtc",O_RDONLY))==-1)
  {
    shmdt(pulse_width);
    shmctl(shmid,IPC_RMID,0);
    close(fdser);
    printf("Cannot open /dev/rtc\n");
    return(-1);
  }

  /* Configure the rtc for periodic interrupts */
  if(ioctl(fdrtc,RTC_IRQP_SET,RTC_RATE)==-1)
  {
    shmdt(pulse_width);
    shmctl(shmid,IPC_RMID,0);
    close(fdser);
    close(fdrtc);
    printf("Cannot set rtc to generate %d interrupts per second\n",RTC_RATE);
    return(-1);
  }

  /* Enable rtc periodic interrupts */
  if(ioctl(fdrtc,RTC_PIE_ON,0)==-1)
  {
    shmdt(pulse_width);
    shmctl(shmid,IPC_RMID,0);
    close(fdser);
    close(fdrtc);
    printf("Cannot enable rtc periodic interrupts\n");
    return(-1);
  }

  /* Start out safely by setting the display to 0 */
  *pulse_width=0;

  /* Spawn the I/O process */
  switch(iopid=fork())
  {
  case -1:		/* Error */
    shmdt(pulse_width);
    shmctl(shmid,IPC_RMID,0);
    close(fdser);
    close(fdrtc);
    printf("Cannot fork the I/O process\n");
    return(-1);
    break;

  case 0:		/* I'm the child */
    /* Attach to the shared memory block in the child's context */
    if((pulse_width=(int *)shmat(shmid,NULL,0))==(int *)-1)
    {
      close(fdser);
      close(fdrtc);
      printf("Cannot attach to the shared memory block in the I/O process\n");
      return(-1);
    }
    return(io_process());

    shmdt(pulse_width);
    close(fdser);
    close(fdrtc);
    break;
  }

  close(fdser);
  close(fdrtc);

  /* Do we calibrate the display ? */
  if(calibrate)
  {
    printf("You're about to calibrate the Duracell display.\n");
    printf("I'm going to ask you to set the bar manually, so I can work out\n");
    printf("how to display values between 0 and 100.\n\n");
    printf("Wait a bit each time to ensure the display stabilizes itself.\n\n");
    printf("Use \"- [ENTER]\" and \"+ [ENTER]\" to set the bar, and\n");
    printf("\"c [ENTER]\" to validate the calibration.\n");

    calibration_points[0]=0;
    for(i=1;i<5;i++)
    {
      printf("Make the bar show approx. %d%% ...\n",i*25);
      while((c=getchar())!='c')
      {
        if(c=='-')
          *pulse_width-=*pulse_width>0?1:0;
        if(c=='+')
          *pulse_width+=*pulse_width<99?1:0;
        if(c=='-' || c=='+')
          printf("Duty cycle: %d%% ... ",*pulse_width);
      }
      calibration_points[i]=*pulse_width;
      printf("Calibration point stored.\n");
    }

    /* Extrapolate the entire calibration table */
    for(i=0;i<4;i++)
      for(j=0;j<25;j++)
        calibration_table[j+i*25]=calibration_points[i]+
		((calibration_points[i+1]-calibration_points[i])*j)/25;

    /* Write out the calibration file */
    if((calfd=fopen(CALIBRATION_FILE,"w"))==NULL)
      printf("Cannot open %s for writing\n",CALIBRATION_FILE);
    else
    {
      for(i=0;i<100;i++)
        fprintf(calfd,"%d\n",calibration_table[i]);
      fclose(calfd);
    }
  }
  else
  {
    /* Zero the calibration table, just in case */
    for(i=0;i<100;i++)
      calibration_table[i]=0;

    /* Read the calibration file */
    if((calfd=fopen(CALIBRATION_FILE,"r"))==NULL)
      printf("Cannot open %s for reading\n",CALIBRATION_FILE);
    else
    {
      for(i=0;i<100;i++)
        fscanf(calfd,"%d",&calibration_table[i]);
      fclose(calfd);

      while(1)
      {
        /* Calculate something resembling the CPU load */
        if((procstatfd=fopen("/proc/stat","r"))!=NULL)
        {
          fscanf(procstatfd,"%s %ld %ld %ld %ld",buf,&l1,&l2,&l3,&l4);
          fclose(procstatfd);

          i=l1-prev_jiffies;
          prev_jiffies=l1;
          i/=2;

          /* Show the load on the display */
          i=i>99?0:i<0?0:i;
          *pulse_width=calibration_table[i];

          sleep(2);
        }
      }
    }
  }

  /* Stop the I/O process */
  *pulse_width=-1;
  usleep(250000);

  shmdt(pulse_width);
  shmctl(shmid,IPC_RMID,0);

  return(0);
}


/* I/O process that actually drives the display */
int io_process(void)
{
  unsigned long rtc_data;
  int line_state=0,new_line_state=0;
  int cnt=0;
  int i;

  while(1)
  {
    /* Wait for the next periodic interrupt */
    switch(read(fdrtc,&rtc_data,sizeof(rtc_data)))
    {
    case 0:
    case -1:		/* We shouldn't be here */
      printf("rtc read error\n");
      return(-1);
      break;

    default:
      if(*pulse_width>=0 && *pulse_width<100)
      {
        if(cnt==0)
          new_line_state=1;
        if(cnt==*pulse_width)
          new_line_state=0;

        /* Do we need to change the state of the control line ? */
        if(new_line_state!=line_state)
        {
          if(ioctl(fdser,TIOCMGET,&i)==-1)
          {
            printf("ioctl (TIOCMGET) error\n");
            return(-1);
          }

          i=new_line_state?i & (~TIOCM_RTS):i | TIOCM_RTS;

          if(ioctl(fdser,TIOCMSET,&i)==-1)
          {
            printf("ioctl (TIOCMSET) error\n");
            return(-1);
          }

          line_state=new_line_state;
        }

        /* Increment the pulse width counter */
        cnt=(cnt+1)%100;
      }
      else
        /* A pulse width not between 0 and 100 means we have to stop */
        return(0);
    }
  }
}
